% * Copyright (c) 2014, Andrew Kae, UMass Amherst
% * All rights reserved.
% *
% * Redistribution and use in source and binary forms, with or without
% * modification, are permitted provided that the following conditions are met:
% *     * Redistributions of source code must retain the above copyright
% *       notice, this list of conditions and the following disclaimer.
% *     * Redistributions in binary form must reproduce the above copyright
% *       notice, this list of conditions and the following disclaimer in the
% *       documentation and/or other materials provided with the distribution.
% *     * Neither the name of the author nor the organization may be used to 
% *       endorse or promote products derived from this software without specific 
% *       prior written permission.
% *
% * THIS SOFTWARE IS PROVIDED BY Andrew Kae ``AS IS'' AND ANY
% * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
% * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
% * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
% * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
% * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

% Code used in the following papers : 
% 
% * Gary B. Huang, Andrew Kae, Carl Doersch, and Erik Learned-Miller
%   Bounding the Probability of Error for High Precision Optical Character Recognition.
%   Journal of Machine Learning Research (JMLR), 2012. 
%
% * Andrew Kae, Gary Huang, Carl Doersch, and Erik Learned-Miller
%   Improving State-of-the-Art OCR through High-Precision Document-Specific Modeling
%   Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition (CVPR), 2010.

% This code is used to extract SIFT features and run the sift_align algorithm as described
% in the papers above.

function multi_class_sift_align (data_root, svm_dir, model_dir, dataset, doc_num)

data_dir = [data_root '/dataset/' dataset];

fprintf('Creating data files for %s %d\n', dataset, doc_num);

%the word box coordinates of each word
page_boxes = [];

%the images themselves
page_images = [];

%Read the image file
full_filename = [data_dir '/doc' num2str(doc_num) '/doc' num2str(doc_num) '.png'];
im_raw = imread(full_filename);

if (length (size(im_raw)) == 3)
    im_raw = im_raw(:,:,1);
end

%for the purposes of the SVM, we should keep the background to be mostly 0
%check if it's binary or 8 bit grayscale
if (mean(double(im_raw(:))) > 100)
    %it's 8 bit grayscale
    im = imcomplement(im_raw);
elseif (mean(double(im_raw(:)))  > 0)
    im = imcomplement(double(im_raw));
else
    im = im_raw;
end

im = double(im);

%normalize the image
im=im-min(im(:)) ;
im=im/max(im(:)) ;

[im_height, im_width] = size(im);

%i think normalizing the values b/w 0 and 1 will lead to better classification
page_images{doc_num} = mat2gray(im);

%Read the segments file
segment_file = [data_dir '/doc' num2str(doc_num) '/prop_info_beta.txt'];

fid = fopen(segment_file);

count = 0;

word_boxes = [];
boxes = [];

while 1
    line = fgetl(fid);
    
    if ~ischar(line)
        break,   end;
    
    %if the line is just a tab, then a new word is starting
    if (isspace(line))
        %increment the counter to indicate a new word has been found
        if (~ isempty(boxes))
            word_boxes{count} = boxes;
        end
        
        boxes = [];
        count = count + 1;
    else
        %read the string on the next line
        segments = textscan(line, '%s %d %d %d %d');
        
        char = segments{1};
        
        top1 = im_height - segments{2};
        bottom1 = im_height - segments{3};
        left1 = segments{4};
        right1 = segments{5};
        
        if (left1 == 0)
            left1 = left1 + 1;
        end
        
        if (right1 > im_width)
            right1 = int32(im_width);
        end
        
        entry = [char, bottom1 top1 left1 right1];
        
        boxes = [boxes; entry];
    end
end

if (~ isempty(boxes))
    word_boxes{count} = boxes;
end

fclose(fid);

page_boxes{doc_num} = word_boxes;

%Read the confident words file
alpha_file = [data_root '/new_alpha/new_alpha_set_' num2str(doc_num)];

[page_nums, words, indices] = textread(alpha_file, '%u %s %u');

%% create this directory if it doesn't exist

if (~exist(svm_dir, 'dir'))
    mkdir(svm_dir);
end

%% Create/Load the glyph map

glyph_map = [];

glyph_dir = sprintf('%s/glyphs/%s', model_dir, dataset);

if ~exist(glyph_dir, 'dir')
    mkdir(glyph_dir);
end

%check if the glyph map already exists
glyph_map_file = sprintf('%s/glyph_map_%d.mat', glyph_dir, doc_num);

if (exist (glyph_map_file))
    %load the file
    load (glyph_map_file, 'glyph_map');
else
    num_indices = length(indices(:,1));
    
    for i=1:num_indices
        word_index = indices(i);
        page_num = page_nums(i);
        
        word = page_boxes{page_num}{word_index};
        
        num_chars = length(word(:,1));
        
        for j=1:num_chars
            %look through the indices and build up the glyph map
            entry = word(j,:);
            
            glyph_map = add_entry(glyph_map, entry, page_num, word_index, j);
        end
    end
    
    save(glyph_map_file, 'glyph_map');
end

fprintf('Loaded glyph map\n');

%% Load Ground Truth Annotations

fprintf('\n');

gold_file =  [data_dir '/doc' num2str(doc_num) '/doc' num2str(doc_num) ...
    '_gold_chars.txt'];

[gold_page_nums, gold_wordids, gold_charids, chars] = textread(gold_file, '%u %u %u %s', 'delimiter', '\t');

gold = [];

%read in the gold annotations
for i=1:length(gold_page_nums)
    gold{gold_page_nums(i)}{gold_wordids(i)}{gold_charids(i)} = chars(i);
end

fprintf('Loaded ground truth annotations.\n');

%% check to make sure the directories to store the SVM files exist.  If they don't, create them.

multi_dir = [svm_dir '/multi_sift_align'];

if (~exist(multi_dir, 'dir'))
    mkdir(multi_dir);
end

%determine the magnification factor for SIFT by computing the average height, width
total_widths = [];
total_heights = [];

num_boxes = length(page_boxes{doc_num});

for b=1:num_boxes
    word_box = page_boxes{doc_num}{b};
    num_chars = length(word_box(:,1));
    
    total_widths = [total_widths; cell2mat(word_box(:,5)) - cell2mat(word_box(:,4))];
    total_heights = [total_heights; cell2mat(word_box(:,3)) - cell2mat(word_box(:,2))];
end

mean_width = mean(total_widths);
mean_height = mean(total_heights);

%SIFT options
%note that this is for both rows and cols for a total of num_spatial_bins^2 bins
num_spatial_bins = 4;

num_orient_bins = 8;

%set the scale factor so that it takes up the whole image
sigma = 1;

%determine the appropriate magnification for this task
magnif = ceil(max([mean_height mean_width]) / num_spatial_bins);

% I've noticed empirically having a magnification smaller than 3 does not help
%make sure magnif >= 3
magnif = max(magnif, 3);

TRAIN_CUTOFF = 5;

%create all the SVM training indices from the clean set data
num_glyphs = size(glyph_map, 1);

SVM_train_cases = [];

mean_dps = [];
training_chars = [];

valid_count = 1;

%% compute the average sift descriptor for each class

fprintf('Computing average sift descriptors\n');

for i=1:num_glyphs
    %only add if there are at least 5 instances of each training class
    instances = glyph_map{i, 2};
    
    num_instances = size(instances,1);
    
    if (num_instances >= TRAIN_CUTOFF)
        training_chars = [training_chars; glyph_map{i,1}];
        SVM_train_cases = [ SVM_train_cases; [valid_count * ones(num_instances,1) instances]];
        
        valid_count = valid_count + 1;
        
        dps = [];
        
        for j=1:num_instances
            patch = page_images{instances(j,1)}(instances(j,4):instances(j,5), instances(j,6):instances(j,7));
            
            [patch_height, patch_width] = size(patch);
            
            dp = siftdescriptor(patch, [round(patch_width/2) - 1; round(patch_height/ 2) - 1; 0], sigma, ...
                'NumSpatialBins', num_spatial_bins, 'Magnif', magnif, 'NumOrientBins', num_orient_bins) ;
            
            dps = [dps; dp'];
        end
        
        mean_dp = mean(dps);
        
        mean_dps = [mean_dps; mean_dp];
    end
end

%% Now write out the SVM files

SVM_test_file = [multi_dir '/SVM_multi_sift_align_test.dat'];
SVM_train_file = [multi_dir '/SVM_multi_sift_align_train.dat'];
Tess_file = [multi_dir '/tess_test'];

gold_labels = [];

SVM_test_cases = [];

%for now, put only what tesseract thinks is one of these characters
p = doc_num;

%try all characters
num_boxes = length(page_boxes{p});

for b=1:num_boxes
    
    word_box = page_boxes{p}{b};
    num_chars = length(word_box(:,1));
    
    for c=1:num_chars
        
        tess_char = word_box{c, 1};
        
        gold_label = cell2mat(gold{p}{b}{c});
        
        %check whether tesseract thinks it's one of the clean chars
        if (ismember(tess_char, training_chars))
            %next check if the gold label is one of these chars
            if (ismember(gold_label, training_chars))
                %make sure it's a single character
                if (length(tess_char) == 1) && (length(gold_label) == 1)
                    %check that the char is not part of the clean set
                    if (~ismember([p b c], SVM_train_cases(:, 2:4), 'rows'))
                        %add this character instance to the test set
                        row = word_box(c, :);
                        
                        [blah, gold_index] = ismember(gold_label, training_chars);
                        [blah, tess_index] = ismember(tess_char, training_chars);
                        
                        SVM_test_cases = [SVM_test_cases; [gold_index tess_index p b c row{2} row{3} row{4} row{5}]];
                    end
                end
            end
        end
    end
end

%% Prepare the datasets to use for the SVM

fprintf('Writing out SVM files.\n');

process_SVM_align_SIFT(SVM_train_cases, SVM_test_cases, page_images, SVM_train_file, SVM_test_file, ...
    Tess_file, mean_height, mean_width, mean_dps);
end